/*                                                                           
**  fi_lib++  --- A fast interval library (Version 2.0)                     
**                                                                  
**  Copyright (C) 2001:                                                        
**                                                     
**  Werner Hofschuster, Walter Kraemer                               
**  Wissenschaftliches Rechnen/Softwaretechnologie (WRSWT)  
**  Universitaet Wuppertal, Germany                                           
**  Michael Lerch, German Tischler, Juergen Wolff von Gudenberg       
**  Institut fuer Informatik                                         
**  Universitaet Wuerzburg, Germany                                           
** 
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/
#include "test/config.hpp"
#include "test/ieee.h"

// -----------------------------------------------------------------------
// Internal functions to compute special numbers
// -----------------------------------------------------------------------
double Double::computeMin() 
{
  a_diee f;
  f.ieee.sign  = 0x0;
  f.ieee.expo  = 0x0;
  f.ieee.mant0 = 0x0;
  f.ieee.mant1 = 0x1;
  return f.f;
}

double Double::computeMinNorm() 
{
  a_diee f;
  f.ieee.sign  = 0x0;
  f.ieee.expo  = 0x1;
  f.ieee.mant0 = 0x0;
  f.ieee.mant1 = 0x0;
  return f.f;
}

double Double::computeMax() 
{
  a_diee f;
  f.ieee.sign  = 0x0;
  f.ieee.expo  = 0x7FE;
  f.ieee.mant0 = 0xFFFFF;
  f.ieee.mant1 = 0xFFFFFFFF;
  return f.f;
}

double Double::computePosInf() 
{
  a_diee f;
  f.ieee.sign  = 0x0;
  f.ieee.expo  = 0x7FF;
  f.ieee.mant0 = 0x0;
  f.ieee.mant1 = 0x0;
  return f.f;
}

double Double::computeNegInf() 
{
  a_diee f;
  f.ieee.sign  = 0x1;
  f.ieee.expo  = 0x7FF;
  f.ieee.mant0 = 0x0;
  f.ieee.mant1 = 0x0;
  
  return f.f;
}

double Double::computeQNaN() 
{
  a_diee f;
  f.ieee.sign  = 0x0;
  f.ieee.expo  = 0x7FF;
  f.ieee.mant0 = 0x80000;
  f.ieee.mant1 = 0x0;
  return f.f;
}

// ----------------------------------------------------------------------- 
// pred/succ table stuff
// -----------------------------------------------------------------------
#ifdef FILIB_PRED_SUCC_TABLES

Double::PredSuccTable::PredSuccTable() 
{
  a_diee f;
  f.ieee.sign  = 0;
  f.ieee.mant0 = 0;
  f.ieee.mant1 = 0;

  ULP = new double[2048];
  for( int i=0; i<2048; i++ ) {
    f.ieee.expo = i;
    ULP[i] = Double::ulp(f.f);
    
    //cout << i << ": \t";
    //Double::bitImage(ULP[i], cout);
    //cout << endl;
  }
}

Double::PredSuccTable::~PredSuccTable() 
{
  delete [] ULP;
}

Double::PredSuccTable Double::psTable;

#endif

// -----------------------------------------------------------------------
// Platform dependent code for FPU initialization
// -----------------------------------------------------------------------

Double::FPUStartUp::FPUStartUp() 
{
#ifdef __i386__
#if defined(__GNUC__) || defined(__KCC)
  // set FPU control to double precision, rounding to nearest, no interrupts
  asm("fldcw %0" : : "m" (0x027F) );
#endif 
#endif // __i386__
}

Double::FPUStartUp startup;

// -----------------------------------------------------------------------
// Instantiation of static members of class Double
// -----------------------------------------------------------------------
double Double::min     = computeMin();  
double Double::minNorm = computeMinNorm();  
double Double::max     = computeMax();
double Double::posInf  = computePosInf();
double Double::negInf  = computeNegInf();
double Double::qNaN    = computeQNaN();

// -----------------------------------------------------------------------
// I/O
// -----------------------------------------------------------------------
void Double::print(double x, std::ostream &os) 
{
  if (x == NEG_INFTY())
    os << "-INF";
  else if (x == POS_INFTY())
    os << "+INF";
  else if (isNaN(x))
    os << "NaN";
  else
    os << x;
}

void Double::bitImage(double d, std::ostream &os) 
{
  basicBitImage(d, os);
  os << std::endl;
}

void Double::basicBitImage(double d, std::ostream &os) 
{
  a_diee f;
  f.f = d;

  int i;

  os << (f.ieee.sign == 1 ? '1' : '0') << ':';
  
  i = 11;
  while (i--)
      os << ((f.ieee.expo & (1 << i)) ? '1' : '0');
  os << ':';
  
  i = 20;
  while (i--)
      os << ((f.ieee.mant0 & (1 << i)) ? '1' : '0');
  
  i = 32;
  while (i--)
      os << ((f.ieee.mant1 & (1 << i)) ? '1' : '0');
}













